#include "SimplyFire.hpp"

const char*  simplyfire_cpp_version(){return "$Revision: 1.1 $";}

/***********************************************
 * Initialisation and fetch results
 ***********************************************/
SimplyFire::SimplyFire(TGlobals& globals,const VarVal& prev_vals,const var_hp_t& v_act_hp, vector<Literal>& unitsout,BitVector<Variable>& flags)
:glob(globals)
//,ass_ifo(assigns)
,units_out(unitsout)
,var_flgs(flags)
,uq_begin(0)
,vals(prev_vals,true)
,clses()
,var_occs('O')
,wtching('S')
,scc(glob,clses,vals,algo_space)
,assy(glob,var_flgs,vals,clses,algo_space,v_act_hp,wtching,units_unpropd,for_deletion)
{}

void SimplyFire::initSimpl(Watching& wtch, const uint32_t& assym_permil,const bool enq_all, const bool enq_lrnts_and_chnd){
	uint32_t sz 		= wtch.litsCount();
	const bool assym	= assym_permil;
	LOG1("start simpl with assym permil " << assym_permil<< " enq_all " << enq_all << " enq_lrnts " << enq_lrnts_and_chnd);

	wtching		.enlarge(sz);
	clses		.ensureLits(sz);
	lit_flgs		.ensureSize(sz,false);
	algo_space	.resize(sz);

	sz >>= 1;
	var_occs	.reserveForTotal(wtch.getTotMem());
	var_occs	.enlarge(sz);
	elim_hp		.insertElementsTo(sz,OccCount());

	FeedbackForm mem_fbk;
	glob.c_info.registerFeedbackForm(mem_fbk);

	WtchEle we(INVALID_LIT,true); Literal ow; ECP ecp; signt_t sig; bool shrnkd;
	for(uint32_t i=0;i!=wtch.litsCount();++i){
		Literal lit(i); const bool lkwn = vals.litKnwn(lit);

		Iter<WtchEle> wi = wtch.getWtchOf(lit);
		while(wi.getNext(we)){

			if(we.isBinary()){
				ow = we.getLit();
				if(!lkwn && lit > ow && !vals.litKnwn(ow)){LOG2("Binary in Init with " << lit.toDIMACS() << " " << ow.toDIMACS());
					clses.addBin(lit,ow);}

			}else{	asrt(wi.hasNext());
				wi.getNext(we);
				const ClsP clp 	= we.getClsP();
				Cls& cls			= glob.getCls(clp);
				ow				= cls.otherWatcher(lit);
				if(lkwn || lit==ow || vals.litKnwn(ow)){
					cls.unsetWatchers(); 						goto SKIP_CLS;}

				if(lit<ow){//see first time

					const int32_t uw_cnt = removeFalseLits(cls,lit,ow,sig,shrnkd);
					//new cls is binary - delete this clause and keep binary
					if(uw_cnt<=0){
						if(uw_cnt==0){LOG2("Binary in Init Shrink"); clses.addBin(lit,ow);}
						cls		.unsetWatchers();				goto SKIP_CLS;}

					mem_fbk.markBorder(clp); //start of clause
					mem_fbk.markBorder(clp+uw_cnt); //last word of clause

					uint32_t lbd	= cls.isLrnt() ? cls.compLBD() : 0;
					uint32_t sz		= uw_cnt + 2;
					bool do_assym	= assym && ((lbd * 100) <= (sz * glob.sets.assym_lbd_cand))
											&&	(glob.randy.nextInt(1000)<assym_permil); //for all fix clses and for those with lbd <= sz/2

					//bool enq_sub	= enq_all || (enq_lrnts_and_chnd && (shrnkd || cls.isLrnt()));
					bool enq_sub	= enq_all || (enq_lrnts_and_chnd && shrnkd);


					ecp = clses.addCls(clp,Literal(cls.uwBeg()->lit),sz,cls.isLrnt(),sig,enq_sub,false,do_assym);
					registerOcc(lit,ow,ecp,cls,cls.isLrnt());

					cls.uwBeg()->lit = ecp.idx; //dirty way to tell the other watcher the address of the wrapping ExtCls

				}else{//see second time
					ecp 			= ECP(cls.uwBeg()->lit);
					ExtCls& ecl 	=  clses.getECls(ecp);
					cls.uwBeg()->lit= ecl.getLit().idx();
					ecl.setLit(lit);
				}

				wtching.add(lit,EWtch(clp,ecp));
			}
			SKIP_CLS:;
		}
	}
	wtch.hardClear();
	glob.c_info.submitFeedback(mem_fbk);

	sanity(clses.asrtBinBothDirs(vals)); sanity(asrtWatchingOK());
}

void SimplyFire::getSimplResult(Watching& wtch){
	sanity(asrtWatchingOK()); sanity(clses.asrtBinBothDirs(vals)); asrt(for_deletion.empty()); asrt(units_unpropd.empty());

	glob.clearLBDDistri();
	glob.stats.cnt_lng_lts 	= 0;
	glob.stats.cnt_lcls		= 0;
	glob.stats.cnt_bin_lts	= 0;

	wtch.enlarge(wtching.listCount());

	FeedbackForm mem_fbk;
	glob.c_info.registerFeedbackForm(mem_fbk);

	Literal ow; EWtch ewtch(ClsP(0),ECP(0)); signt_t sig; bool shrnkd;
	for(uint32_t i=0;i!=wtching.listCount();++i){
		Literal lit(i); const bool kwn = vals.litKnwn(lit);

		if(!kwn){
			Iter<Literal> bs	= clses.getBinsOf(lit);
			while(bs.getNext(ow)){
				if(!vals.litKnwn(ow)){wtch.clsCreateBinPart(lit,ow);cnt_stat(bin_lts,1); }
			}
		}

		Iter<EWtch> ew = wtching.getIter(lit);
		while(ew.getNext(ewtch)){
			ExtCls& ecl 	= clses.getECls(ewtch.extp);
			if(ecl.isWaste()||kwn){ecl.doWasteWasAlve();		goto SKIP_CLS;}

			{Cls& cls	= glob.getCls(ewtch.clp);
			ow			= cls.otherWatcher(lit); asrt(lit!=ow); asrt(ecl.getLit()==lit || ecl.getLit()==ow);

			if(lit<ow){ //see first time
				LOG2("Check cls " << ewtch.clp.idx() << " with watchers " << lit.toDIMACS() << " and " << ow.toDIMACS());

				if(vals.litKnwn(ow)){asrt(vals.litTrue(ow));
						ecl.doWaste(); 							goto SKIP_CLS;}

				const int32_t uw_cnt = removeFalseLits(cls,lit,ow,sig,shrnkd); asrt(!shrnkd);
				if(uw_cnt <= 0){
					if(uw_cnt==0){wtch.clsCreateBinary(lit,ow); cnt_stat(bin_lts,2);}
					ecl.doWaste();								goto SKIP_CLS;}

				mem_fbk.markBorder(ewtch.clp); //start of clause
				mem_fbk.markBorder(ewtch.clp+uw_cnt); //last word of clause
				cls.setLrnt(ecl.isLearnt());

				if(cls.isLrnt()){glob.registrLBD(cls.compLBD());}

				cnt_stat(lng_lts,(uw_cnt+2));
				cnt_stat(lcls,1);
			}}
			wtch.clsAddWtched(lit,ow,ewtch.clp);
			LOG2("Watch cls " << ewtch.clp.idx() << " from lit " << lit.toDIMACS());
			SKIP_CLS:;
		}
	}

	wtching.hardClear();
	glob.c_info.submitFeedback(mem_fbk);
}


/*******************************
 * Var elimination and blocked clauses
 *******************************/
struct ValPrinter{
	uint32_t operator()(const OccCount& o) const {return o.product();}
};

bool SimplyFire::elim(const bool blkd_chk, uint32_t& progress){
	progress = 0; T_ON(elimchk);
	const uint32_t max_product 	= (glob.sets.eli_max_factor * glob.sets.eli_max_factor) + glob.sets.eli_grow;
	const uint32_t max_lits		= glob.sets.eli_max_lits;
	const uint32_t max_l_split 	= blkd_chk ? max32 : max_lits;
	LOG2("Start elim; max_prod " << max_product << " max_lits " << max_lits << " max lits for split " << max_l_split << " blockd check " << blkd_chk);
	bool res 					= true;
	vector<FaktCls> split[2];
	ElimProcut eli_prod;
	//ValPrinter printer;

	while(!elim_hp.empty()){
		asrt(units_unpropd.empty()); asrt(for_deletion.empty()); asrt(eli_prod.asrtAllEmpty()); //sanity(elim_hp.asrtHeapProperty());
		sanity(asrtOccCount());

		//elim_hp.printHeap(printer);

		Variable var 		= elim_hp.removeTop();
		if(vals.varKnwn(var) || glob.isElim(var)){ 		LOG2("Got unusable var" << var.toDIMACS());
			continue;}

														LOG2("Check of variable " << var.toDIMACS());
		split[0].clear(); split[1].clear();
		OccCount cnt = elim_hp.getValue(var); 			asrt(elim_hp.empty() || cnt.product() <= elim_hp.getTop().product());
		if(cnt.product() > max_product && !blkd_chk){
			elim_hp.clear();
			break;
		}

		uint32_t lits	= splitFixClses(var,split,max_l_split); LOG2("Split result: false cnt: " << split[0].size() << " true cnt: " << split[1].size());

		const bool pos_sec		= split[0].size() < split[1].size();
		vector<FaktCls>& fakt1	= split[!pos_sec];
		vector<FaktCls>& fakt2	= split[pos_sec]; asrt(fakt1.size() <= fakt2.size());
		const Literal f2Lit		= var.toLit(pos_sec);
		const Literal f1Lit		= f2Lit.flipped();

		if(fakt1.size() > 64){ LOG2("Less occurs are still: " << fakt1.size() << " quit now");
			//consider that elim_hp does not known about binary occurrences
			elim_hp.clear();
			break;
		}

		if(fakt1.empty()){LOG1("Pure Literal found: " << f2Lit.toDIMACS());
			finallyEliminate(var);
			continue;
		}

		if(lits <= max_lits) {LOG2("To many lits " << lits << " in var elim of " << var.toDIMACS());

			res = elimVarIf(var,eli_prod,f1Lit,fakt1,f2Lit,fakt2,max_product,true);}

		else if(blkd_chk){ asrt(lits != max32); //split was completed

			res = blockedCheckVar(var,eli_prod,f1Lit,fakt1,f2Lit,fakt2);
		}

		if(!res){T_OFF(elimchk);
			return glob.setInvalid("Empty clause in var elimination");}


		++progress;

#ifdef PRINT_ELIM_SUB
		if(glob.sets.print_console){
			printf("Vars to check: %d \r",elim_hp.size());}
#endif

	}

	T_OFF(elimchk);
	return true;
}

uint32_t SimplyFire::splitBinsHlpr(const Literal& a, vector<FaktCls>& add){
	Iter<Literal> it = clses.getBinsOf(a); Literal lit;
	const uint32_t s = add.size();
	while(it.getNext(lit)){ add.push_back(FaktCls(lit)); asrt(vals.litUkwn(lit)); asrt(!glob.isElim(lit));}
	return add.size() - s; //only count literals not double
}


//split has size 2 -> one vector for negative occurrences, one for positive
uint32_t SimplyFire::splitFixClses(const Variable& var,vector<FaktCls> split[],const uint32_t& max_lits){
	asrt(!vals.varKnwn(var)); asrt(!glob.isElim(var)); sanity(asrtOccCountOf(var)); asrt(split[0].empty()); asrt(split[1].empty());
	LOG2("Split clauses of var " << var.toDIMACS());

	OccCls oc(ECP(0),false);
	uint32_t lits	= 0;
	WIter<OccCls> oi = var_occs.getWIter(var);

	while(oi.getNext(oc)){
		ExtCls& ec = clses.getECls(oc.getECP());
		if(ec.isWaste()){		continue;}
		oi.keep(oc);
		if(ec.isLearnt()){		continue;}

		lits 					+= ec.size();
		split[oc.isPosOcc()]	.push_back(FaktCls(oc.getECP(),ec.size())); //asrt(asrtPrintCls(oc.getECP(),"split cls to ",oc.isPosOcc()));
	}
	var_occs.finish(var,oi);
	if(lits > max_lits){ 		LOG2("Early break of splitFixClses max was " << max_lits << " we have " << lits);
		return max32;}

	lits 	+= 	splitBinsHlpr(var.toNegLit(),split[0])
			+	splitBinsHlpr(var.toPosLit(),split[1]);

	return lits;
}

uint64_t SimplyFire::initFakt1Lits(const vector<FaktCls>& fakt1, vector<Literal>& cont){
	asrt(fakt1.size() <= 64); sanity(lit_flgs.allCleared());
	LOG2("Init lits of first factor clauses");

	uint64_t all = 0;uint64_t msk = 1;
	for(uint32_t c=0;c!=fakt1.size();++c,msk<<=1){ asrt(msk!=0); asrt(powOf2(msk));
		all	|= msk;

		if(fakt1[c].isBin()){
			setMsk(fakt1[c].getLit(),msk,cont); LOG2("Set mask for bin " << fakt1[c].getLit().toDIMACS() << ": " << msk);
		}else{
			ExtCls& ec 	= clses.getECls(fakt1[c].getECP()); asrt(!ec.isLearnt()); asrt(!ec.isWaste());
			setMsk(ec.getLit(),msk,cont);
			Cls& cls	= glob.getCls(ec.getClp());
			setMsk(cls.otherWatcher(ec.getLit()),msk,cont);
			ClsLit* clit= cls.uwBeg();
			do{
				setMsk(Literal(clit->lit),msk,cont);
			}while((clit++)->has_next);
			LOG2("Set mask for cls " << fakt1[c].getECP().idx << ": " << msk); //asrt(asrtPrintCls(fakt1[c].getECP()," masked cls ",msk));
		}
	}
	LOG2("Mask of all clauses in first factor is " << all);
	return all;
}

void  SimplyFire::addFaktIf(const Literal& lit,uint64_t& add_cands,uint64_t& no_tauts, vector<Literal> fakt1[], const Variable& ignore){
	if(lit.toVar()==ignore) return;
	LOG2("Mask when adding literal " << lit.toDIMACS() << " add_cands: " << add_cands << " no tauts: " << no_tauts);

	uint64_t taut 		= getMsk(lit.flipped()) & no_tauts;	//all positions p are marked where fakt1[p-1] is tautological
	no_tauts			&= ~taut;
	add_cands			&= no_tauts;
	uint64_t add 		= (~getMsk(lit)) & add_cands; //all positions p are marked where lit has to be added in fakt1[p-1]


	LOG2("Masks new. add_cands: " << add_cands << " lit msk: " << getMsk(lit) << " add: " << add);
	LOG2("Mask of flipped: " << getMsk(lit.flipped()) << " taut:  " << taut << " no tauts " << no_tauts);
	for(uint32_t i=0;add || taut;++i,add>>=1,taut>>=1){ asrt(((add&1) ==0 ) || (taut&1)==0);
		if(add&1){ LOG2("Add to cross prod bitmp : " << lit.toDIMACS() << " at " << i);
			fakt1[i].push_back(lit);}
		else if(taut&1){LOG2("Taut cross prod bitmp : " << lit.flipped().toDIMACS() << " at " << i);
			fakt1[i].clear(); fakt1[i].push_back(TRUE_LIT);
		}
	}
}

void SimplyFire::addFakt2Lits(const vector<FaktCls>& fakt2,const uint64_t& all_clses_msk, vector<vector<Literal> >& cross, const uint32_t& f1sz, const Variable& ignore){
	Literal lit;
	for(uint32_t f2=0;f2!=fakt2.size();++f2){
		vector<Literal>* row = getStartForFakt2(cross,f1sz,f2);
		uint64_t no_tauts	= max64; //all fakt1 clauses that are not yet tautological with fakt2[f2]
		uint64_t add_to 	= all_clses_msk;

		if(fakt2[f2].isBin()){
			addFaktIf(fakt2[f2].getLit(),add_to,no_tauts,row,ignore);
		}else{
			ExtCls& ec	= clses.getECls(fakt2[f2].getECP());
			addFaktIf(ec.getLit(),add_to,no_tauts,row,ignore);
			Cls& cls	= glob.getCls(ec.getClp());
			addFaktIf(cls.otherWatcher(ec.getLit()),add_to,no_tauts,row,ignore);
			ClsLit* clit= cls.uwBeg();
			do{
				if(no_tauts==0){LOG2("All tauts "); return;}
				addFaktIf(Literal(clit->lit),add_to,no_tauts,row,ignore);
			}while((clit++)->has_next);
		}
	}
}

void SimplyFire::bwSelf(FaktCls& fcls, const Literal& rem){
	inc_stat(eliself,1);
	if(fcls.isBin()){ LOG2("Remove lit " << rem.toDIMACS() << " of fact cls causes unit " << fcls.getLit().toDIMACS());
		units_unpropd.push_back(fcls.getLit());
	}else{
		ECP ecp 	= fcls.getECP();
		ExtCls& ec 	= clses.getECls(ecp);
		removeLit(ecp,rem);
		bool revive	= 	ec.doWasteWasAlve();
		cleanOccs(rem.toVar());
		if(revive){		ec.undoWaste();}
		LOG2("Remove lit " << rem.toDIMACS() << " of fact cls " << fcls.getECP().idx);
	}
	sanity(asrtOccCount());
}

bool SimplyFire::elimVarIf(const Variable& var, ElimProcut& eli_prod, const Literal& f1Lit, vector<FaktCls>& fakt1, const Literal& f2Lit,vector<FaktCls>& fakt2,
				const uint32_t& max_prod, bool assym){

	LOG2("Start elim check of var really " << var.toDIMACS());
	asrt(!vals.varKnwn(var)); asrt(!glob.isElim(var));	sanity(lit_flgs.allCleared()); sanity(asrtOccCount()); asrt(eli_prod.asrtAllEmpty());
	asrt(units_unpropd.empty()); asrt(clean_lits.empty()); asrt(fakt1.size() <= fakt2.size()); asrt(fakt1.size() <=64);

	const uint32_t f1sz		= fakt1.size();
	const uint32_t f2sz		= fakt2.size(); LOG2("Smaller size is " << f1sz << " bigger fact is " << f2sz);
	asrt(f1sz>0); asrt(f2sz>0);

	//mark all literals of clses in outr
	uint64_t all = initFakt1Lits(fakt1,clean_lits); 	asrt(!lit_flgs.isSet(f2Lit));
	lit_flgs.clear(f1Lit);	//setMsk(f2Lit,all,clean); //ensure that outr lit is not set, and inner lit is already contained in all cross products

	eli_prod.ensureElimClsProd(f1sz*f2sz);
	addFakt2Lits(fakt2,all,eli_prod.elim_cls_prod,f1sz,var);

	//clear marks
	for(uint32_t i=clean_lits.size();i--;){lit_flgs.carelessClear(clean_lits[i]);} sanity(lit_flgs.allCleared());
	clean_lits.clear();

	vector<FiniProd> prods;
	uint32_t f1_used = 0;
	uint32_t f2_used = 0;
	for(uint32_t f2=0;f2!=f2sz;++f2){
		const uint32_t f2_cnt		 = fakt2[f2].getSz(); //lits in second clause without var
		vector<Literal>* allF1WithF2 = getStartForFakt2(eli_prod.elim_cls_prod,f1sz,f2); //all product parts to f2
		for(uint32_t f1=0;f1!=f1sz;++f1){
			vector<Literal>& prod_prt	= allF1WithF2[f1];
			uint32_t f1_cnt				= fakt1[f1].getSz();
			uint32_t psz		  		= prod_prt.size() + f1_cnt;


			/*LOG2("First product cls with var : " << var.toDIMACS());
			if(fakt1[f1].isBin()){
				LOG2("Bin: " << f1Lit.toDIMACS() << " " << fakt1[f1].getLit().toDIMACS());
			}else{
				asrtPrintCls(fakt1[f1].getECP(),"Fakt 1 red sz: ",fakt1[f1].getSz());
			}

			LOG2("Second product cls with var : " << var.toDIMACS());
			if(fakt2[f2].isBin()){
				LOG2("Bin: " << f2Lit.toDIMACS() << " " << fakt2[f2].getLit().toDIMACS());
			}else{
				asrtPrintCls(fakt2[f2].getECP(),"Fakt 2 red sz: ",fakt2[f2].getSz());
			}*/


			if(!prod_prt.empty() && prod_prt[0]==TRUE_LIT){
				LOG2("Product of f1 " << f1 << " and f2 " << f2 << " is taut");
				prod_prt.clear();
				continue;
			}

			if(psz <= f1_cnt || psz <= f2_cnt){
				LOG2("Self BW Subsumption of one fakt, f1: " << f1_cnt << " and f2: " << f2_cnt << " and prod: " << psz);

				if(psz <= f1_cnt){bwSelf(fakt1[f1],f1Lit);}
				if(psz <= f2_cnt){bwSelf(fakt2[f2],f2Lit);}

				elim_hp	.restoreElementIf(var); asrt(!glob.isElim(var));
				eli_prod.elim_cls_prod.clear();
				return propagateUnpropdUnits();
			}

			asrt(!prod_prt.empty());
			f1_used += fakt1[f1].setUsdNew();
			f2_used += fakt2[f2].setUsdNew();
			prods.push_back(FiniProd(fakt1[f1],&prod_prt));
		}

		//if(prods.size() > max_prods) break;
	}
	asrt(f1_used <= f1sz); asrt(f2_used <= f2sz);


	if(prods.size()>max_prod){
		eli_prod.elim_cls_prod.clear();

		inc_stat(elifail,1); LOG2("Elim check failed of " << var.toDIMACS());
		if(f1_used != f1sz){elimBlockdIf(fakt1,f1Lit);}
		if(f2_used != f2sz){elimBlockdIf(fakt2,f2Lit);}
		delClses();
		return true;
	}

	//add clauses
	//std::sort(prods.begin(),prods.end());
	for(uint32_t i=0; i!=prods.size();++i){
		FiniProd& pr 			= prods[i];
		vector<Literal>& lts 	= *pr.sec_part;
		if(pr.fctcls.isBin()){
			lts.push_back(pr.fctcls.getLit());
		}else{
			addAllLitsBut(pr.fctcls.getECP(),var,lts);}

		if(lts.size() == 2){
			clses.addBin(lts[0],lts[1]);
		}else{
			prods[i].setProd( //only needed for assym checks later
					newCls(lts.begin().base(),lts.size(),false,0,true,true),lts.size());
			asrt(!clses.getECls(prods[i].fctcls.getECP()).isWaste());
		}
		lts.clear();
	}

	sanity(clses.asrtBinBothDirs(vals)); sanity(eli_prod.asrtAllEmpty());

	finallyEliminate(var);

	return  assym ? productAssymCheck(prods) : true ;
}


void SimplyFire::elimBins(const Literal& lit,vector<Literal>& out){
	Literal ol;
	Iter<Literal> bit = clses.getBinsOf(lit);
	while(bit.getNext(ol)){
		glob.elim_clses.remember(lit,ol,VAR_ELIM);
		if(lit_flgs.setFlagNew(ol)){out.push_back(ol);}
	}
	clses.clearBinsPartOf(lit);
}



void SimplyFire::elimBlockdIf(vector<FaktCls>& cands, const Literal& anc){
	asrt(!glob.isElim(anc.toVar()));
	bool did = false;

	for(uint32_t i=cands.size();i--;){
		FaktCls& cl = cands[i];
		if(cl.isUsed()) continue;
		did = true;
		if(cl.isBin()){
			Literal lit = cl.getLit(); LOG2("Elim blocked binary clause " << anc.toDIMACS() << " " << lit.toDIMACS());
			glob.elim_clses.remember(anc,lit,BLKD_CLS); asrt(!glob.isElim(lit.toVar()));
			clses.removeBinPart(lit,anc);
			clses.removeBinPart(anc,lit);
			elim_hp.restoreElementIf(lit.toVar());
		}else{
			inc_stat(elimblkd,1);
			ExtCls& ec = clses.getECls(cl.getECP()); sanity(asrtPrintCls(cl.getECP()," remember blocked for ", anc.toDIMACS()));
			glob.elim_clses.rememberBlkd(ec.getLit(), glob.getCls(ec.getClp()), ec.size(),anc);
			for_deletion.push_back(Subsumd(cl.getECP(),ECP(max32)));
		}
	}
	if(did){elim_hp.restoreElementIf(anc.toVar());}
}

void	SimplyFire::finallyEliminate(const Variable& var){
	asrt(!glob.isElim(var)); asrt(clean_lits.empty()); LOG2("Finally elim now var " << var.toDIMACS());

	//ensure that variable gets a value at model computation
	glob.elim_clses	.remember(var);
	glob			.setElim(var); inc_stat(elim,1);

	//remove long clause
	Iter<OccCls> iter = var_occs.getIter(var); OccCls oc(ECP(0),false);
	while(iter.getNext(oc)){
		ExtCls& ec = clses.getECls(oc.getECP());
		if(!ec.isLearnt()){ sanity(asrtPrintCls(oc.getECP()," remember elim var for ", var.toLit(oc.isPosOcc()).toDIMACS()));
			glob.elim_clses.remember(ec.getLit(),glob.getCls(ec.getClp()),ec.size(),var.toLit(oc.isPosOcc()),VAR_ELIM);}

		for_deletion.push_back(Subsumd(oc.getECP(),ECP(max32))); sanity(asrtPrintCls(oc.getECP(),"elimination of var ", var.toDIMACS()));
	}
	var_occs.clear(var);

	delClses();

	elim_hp			.removeIfContained(var);

	//clean bins
	elimBins(var.toLit(false),clean_lits);
	elimBins(var.toLit(true),clean_lits);
	cleanBinsAndFlgs(clean_lits); sanity(clses.asrtBinBothDirs(vals));
	clean_lits.clear();

	LOG2("Finally eliminated var " << var.toDIMACS());
}


bool   SimplyFire::blockedCheckVar(const Variable& var,ElimProcut& eli_prod, const Literal& f1Lit, vector<FaktCls>& fakt1, const Literal& f2Lit,vector<FaktCls>& fakt2){

	LOG2("Start blocked check of var " << var.toDIMACS());
	asrt(!vals.varKnwn(var)); asrt(!glob.isElim(var));	sanity(lit_flgs.allCleared()); sanity(asrtOccCount());
	asrt(units_unpropd.empty()); asrt(clean_lits.empty()); asrt(fakt1.size() <= fakt2.size()); asrt(fakt1.size() <=64); sanity(eli_prod.asrtAllEmpty());

	const uint32_t f1sz		= fakt1.size();
	const uint32_t f2sz		= fakt2.size(); LOG2("Smaller size is " << f1sz << " bigger fact is " << f2sz);
	asrt(f1sz>0); asrt(f2sz>0);

	//mark all literals of clses in outr
	uint64_t all = initFakt1Lits(fakt1,clean_lits); 	asrt(!lit_flgs.isSet(f2Lit));
	lit_flgs.clear(f1Lit);	//setMsk(f2Lit,all,clean); //ensure that outr lit is not set, and inner lit is already contained in all cross products


	eli_prod.ensureBlkdClsProd(f1sz*f2sz);
	initFakt2Lits(fakt2,all,eli_prod.blkd_cls_prod,f1sz,var);

	//clear marks
	for(uint32_t i=clean_lits.size();i--;){lit_flgs.carelessClear(clean_lits[i]);} sanity(lit_flgs.allCleared());
	clean_lits.clear();

	uint32_t f1_used = 0;
	uint32_t f2_used = 0;
	for(uint32_t f2=0;f2!=f2sz;++f2){
		const uint32_t f1_strt		= getStrtForFakt2(f1sz,f2); //all product parts to f2

		for(uint32_t f1=0;f1!=f1sz;++f1){
			const uint32_t cidx = f1_strt + f1;


			if(eli_prod.blkd_cls_prod.firstSet(cidx)){ //taut
				LOG2("Product of f1 " << f1 << " and f2 " << f2 << " is taut");

			}else if(eli_prod.blkd_cls_prod.secSet(cidx)){ //product is not tautological and contains at least one literal of the second factor

				f1_used += fakt1[f1].setUsdNew();
				f2_used	+= fakt2[f2].setUsdNew();

				LOG2("Count product. fiused " << f1_used << " and f2used " << f2_used );

			}else{ //product does not contain any literal of the the second factor, so factor1 is subsumed by the product

				LOG2("Product of f1 " << f1 << " and f2 " << f2 << " is selfsubsuming of f1");
				bwSelf(fakt1[f1],f1Lit);

				elim_hp	.restoreElementIf(var); asrt(!glob.isElim(var));
				eli_prod.blkd_cls_prod.hardClear();
				return propagateUnpropdUnits();
			}
			eli_prod.blkd_cls_prod.clearBoth(cidx);
		}
	}
	asrt(f1_used <= f1sz); asrt(f2_used <= f2sz); asrt(eli_prod.asrtAllEmpty());

	if(f1_used != f1sz){elimBlockdIf(fakt1,f1Lit);}
	if(f2_used != f2sz){elimBlockdIf(fakt2,f2Lit);}
	delClses();

	return true;
}


void  SimplyFire::initFakt2If(const Literal& lit,uint64_t& add_cands,uint64_t& no_tauts,TwoBitsVec<uint32_t>& cross, uint32_t f1_strt, const Variable& ignore){
	if(lit.toVar()==ignore) return;
	LOG2("Mask when adding literal " << lit.toDIMACS() << " add_cands: " << add_cands << " no tauts " << no_tauts);

	uint64_t taut 		= getMsk(lit.flipped()) & no_tauts;	//all positions p are marked where fakt1[p-1] is tautological
	no_tauts			&= ~taut;
	add_cands			&= no_tauts;
	uint64_t add 		= (~getMsk(lit)) & add_cands; //all positions p are marked where lit has to be added in fakt1[p-1]
	add_cands			&= ~add; asrt((add&taut)==0); asrt((add_cands & ~no_tauts) == 0);


	LOG2("More Masks.  add_cands: " << add_cands << " lit msk: " << getMsk(lit) << " add: " << add);
	LOG2("Mask of flipped: " << getMsk(lit.flipped()) << " taut:  " << taut << " no tauts " << no_tauts);
	for(;add || taut;++f1_strt,add>>=1,taut>>=1){ asrt(((add&1) ==0 ) || (taut&1)==0);
		LOG2("Check blocked elim entry at " <<f1_strt);

		if(add&1){ LOG2("Add to cross prod bitmp : " << lit.toDIMACS()); asrt(!cross.secSet(f1_strt));
			cross.setSec(f1_strt);}
		else if(taut&1){LOG2("Taut cross prod bitmp : " << lit.flipped().toDIMACS() ); 		asrt(!cross.firstSet(f1_strt));
			cross.setFirst(f1_strt);
		}
	}
}

void SimplyFire::initFakt2Lits(const vector<FaktCls>& fakt2,const uint64_t& all_clses_msk, TwoBitsVec<uint32_t>& cross, const uint32_t& f1sz, const Variable& ignore){
	Literal lit; sanity(cross.allCleared());

	for(uint32_t f2=0;f2!=fakt2.size();++f2){
		const uint32_t row 	= getStrtForFakt2(f1sz,f2);
		uint64_t no_tauts	= max64; //all fakt1 clauses that are not yet tautological with fakt2[f2]
		uint64_t add_to 	= all_clses_msk;

		if(fakt2[f2].isBin()){
			initFakt2If(fakt2[f2].getLit(),add_to,no_tauts,cross,row,ignore);
		}else{
			ExtCls& ec	= clses.getECls(fakt2[f2].getECP());
			initFakt2If(ec.getLit(),add_to,no_tauts,cross,row,ignore);
			Cls& cls	= glob.getCls(ec.getClp());
			initFakt2If(cls.otherWatcher(ec.getLit()),add_to,no_tauts,cross,row,ignore);
			ClsLit* clit= cls.uwBeg();
			do{
				if(no_tauts==0){LOG2("All tauts "); return;}
				initFakt2If(Literal(clit->lit),add_to,no_tauts,cross,row,ignore);
			}while((clit++)->has_next);
		}
	}
}




/**********************************************
 * Assym checks of clauses
 **********************************************/
bool SimplyFire::assymChecks(const bool blkd_chk){
	T_ON(assymchk);	sanity(asrtWatchingOK());

	for(uint32_t a=0;a < clses.maxClses() ;++a){ECP	ecp(a);
		ExtCls& ec = clses.getECls(ecp);
		if(ec.isWaste() || !ec.shallAssymChd()){continue;}

		uint32_t nsz 	= assy.shrinkyWinky(ECP(a)); sanity(asrtWatchingOK());

		ec.setAssym(false);
		if(nsz != max32){ asrt(!ec.isWaste()); asrt(nsz<ec.size()); asrt(nsz>2);
			keepMarkdVars(ecp,nsz);
		}

		delClses();

		if(!propagateUnpropdUnits()){ T_OFF(assymchk);
			return glob.setInvalid("Bad units after assym check");}

	}
	sanity(asrtWatchingOK());
	T_OFF(assymchk);

	return elimAndSubsum(blkd_chk);
}

bool SimplyFire::productAssymCheck(vector<FiniProd>& prods){
	T_OFF(elimchk); T_ON(assymchk);

	for(uint32_t i=0;i!= prods.size();++i){
		FiniProd& pr = prods[i];
		if(pr.fctcls.isBin()) continue;

		ECP ecp		= pr.fctcls.getECP();
		ExtCls ec 	= clses.getECls(ecp);
		if(ec.isWaste()) continue;

		uint32_t nsz 	= assy.shrinkyWinky(ecp); sanity(asrtWatchingOK());

		ec.setAssym(false);
		if(nsz != max32){ asrt(nsz > 2); asrt(!ec.isWaste());
			keepMarkdVars(ecp,nsz);
		}

		delClses();

		if(!propagateUnpropdUnits()){ T_OFF(assymchk); T_ON(elimchk);
			return glob.setInvalid("Bad units after assym check");}
	}

	T_OFF(assymchk); T_ON(elimchk);
	return true;
}

void SimplyFire::keepMrkdAux(const Literal& lit, uint32_t& nsz, Literal nw[], ClsLit*& cp,signt_t& sig){

	if(var_flgs.isSet(lit.toVar())){
		var_flgs	.clear(lit.toVar());
		sig 		|= ExtCls::getHash(lit);
		if(nsz<2){	nw[nsz]	=lit;}
		else{		(cp++)->lit = lit.idx();}
		++nsz;
	}else{
		cleanOccs(lit.toVar());
	}
}

//#define keep_if(x){if(var_flgs.isSet(var=x.toVar())){nw[nsz++]=wt[0];var_flgs.clear(wt[0].toVar());}}
void SimplyFire::keepMarkdVars(ECP& ecp, const uint32_t& sz){
	sanity(asrtWatchingOK()); sanity(asrtOccCount()); asrt(sz>2); sanity(asrtPrintCls(ecp,"keep some ", sz));

	ExtCls& ec = clses.getECls(ecp); asrt(!ec.isWaste());
	Cls& cls	= glob.getCls(ec.getClp());
	ec			.doWaste();


	Literal wt	[2];
	bool p 		= !var_flgs.isSet(ec.getLit().toVar()); //if only one watcher is kept, let it be at pos wt[0]
	wt[p]		= ec.getLit();
	wt[!p] 		= cls.otherWatcher(wt[p]); asrt(wt[0] !=wt[1]); asrt(var_flgs.isSet(wt[0].toVar())|| !var_flgs.isSet(wt[1].toVar()));

	ClsLit* clit	= cls.uwBeg();
	ClsLit* cp		= clit;

	Literal nw		[2];
	uint32_t nsz	= 0;
	signt_t sig 	= 0;
	keepMrkdAux(wt[0],nsz,nw,cp,sig);
	keepMrkdAux(wt[1],nsz,nw,cp,sig); asrt(cp==clit);

	do{
		keepMrkdAux(Literal(clit->lit),nsz,nw,cp,sig);
	}while((clit++)->has_next);

	asrt(cp>cls.uwBeg()); asrt(cp-cls.uwBeg()== sz-2); asrt(nsz==sz); sanity(var_flgs.allCleared()); asrt(wt[0] != nw[1]); asrt(wt[1] != nw[0]);

	(--cp)	->has_next 	= 0;
	cls		.setWatchers(nw[0],nw[1]);
	ec		.setLit(nw[0]);
	ec		.setSigSize(nsz,sig);


	if(wt[0] != nw[0]){
		cleanWtches(wt[0]);
		wtching.add(nw[0],EWtch(ec.getClp(),ecp));}

	if(wt[1] != nw[1]){
		cleanWtches(wt[1]);
		wtching.add(nw[1],EWtch(ec.getClp(),ecp));}


	ec		.undoWaste();
	clses	.enqSubTest(ecp); sanity(asrtWatchingOK()); sanity(asrtOccCount());
}



/**********************************************
 * Replace equal variables
 **********************************************/
bool SimplyFire::replaceSCCs(){
	sanity(asrtWatchingOK()); sanity(clses.asrtBinBothDirs(vals)); sanity(asrtOccCount());

	T_ON(scc);
	list<list<Literal> > 	sccs;
	clses.binsUpdated();//reset internal trigger to watch for new clauses
	scc.init();

	do{	asrt(sccs.empty()); asrt(units_unpropd.empty());

		scc.computeSCCs(sccs,units_unpropd); sanity(asrtBinNoElims());

		if(!propagateUnpropdUnits()){T_OFF(scc);
			return glob.setInvalid("Bad units after scc computation");}

		asrt(units_unpropd.empty()); sanity(asrtWatchingOK()); sanity(clses.asrtBinBothDirs(vals)); sanity(clses.asrtNoMrkdClses()); sanity(asrtBinNoElims());

		if(!replaceSCCs(scc, sccs,units_unpropd)){T_OFF(scc);
			return glob.setInvalid("Bad scc replace");}

		if(!propagateUnpropdUnits()){T_OFF(scc);
			return glob.setInvalid("Bad units after scc replace");}

	}while(clses.binsUpdated());

	asrt(units_unpropd.empty()); sanity(asrtWatchingOK());  sanity(asrtOccCount());
	sanity(clses.asrtBinBothDirs(vals)); sanity(clses.asrtNoMrkdClses());

	T_OFF(scc);
	return true;
}

void SimplyFire::updateBins(SCC& scc, const Literal& lit,vector<Literal>& units_out){
	asrt(lit_flgs.allCleared()); LOG2("Update bins of " << lit.toDIMACS());
	updateBinsDontClean(scc,lit,units_out);
	if(lit_flgs.isSet(lit)){units_out.push_back(lit);LOG1("Found unit " << lit.toDIMACS());}

	lit_flgs.carelessClear(lit.flipped());
	Iter<Literal> it = clses.getBinsOf(lit); Literal obin;
	while(it.getNext(obin)){lit_flgs.carelessClear(obin);}
	asrt(lit_flgs.allCleared());
}

void SimplyFire::updateBinsDontClean(SCC& scc, const Literal& lit,vector<Literal>& units_out){
	Literal obin; LOG2("Clean bins  " << lit.toDIMACS());
	WIter<Literal> wi = clses.getWBinsOf(lit); lit_flgs.setFlag(lit.flipped());
	while(wi.getNext(obin)){
		obin = glob.isElim(obin) ? scc.getTmpLit(obin) : obin;
		if(lit_flgs.setFlagNew(obin)){ wi.keep(obin); LOG2("Keep bin  " << obin.toDIMACS());}
		if(lit_flgs.isSet(obin.flipped())){LOG1("Got unit in updateBins " << lit.toDIMACS());
			units_out.push_back(lit);}
	}
	clses.finishBins(lit,wi);
}



void SimplyFire::mergeBinLists(SCC& scc,const Literal& survive, list<Literal>& die, const bool first,vector<Literal>& units_out,vector<Variable>& update){
	asrt(lit_flgs.allCleared()); Literal rep,obin; OccCls oc(ECP(0),false);
	LOG2("Merge lits of  " << survive.toDIMACS());

	//clean binaries first
	updateBinsDontClean(scc,survive,units_out);

	for(list<Literal>::iterator iter=die.begin();iter != die.end(); ++iter){
		rep = first ? *iter : iter->flipped();	LOG2("Append lits of " << rep.toDIMACS());
		Iter<Literal> it = clses.getBinsOf(rep);
		while(it.getNext(obin)){
			obin = glob.isElim(obin) ? scc.getTmpLit(obin) : obin; LOG2("Check other " << obin.toDIMACS());
			if(lit_flgs.isSet(obin.flipped())){	LOG1("Got unit in mergeBinLists " << survive.toDIMACS());
				units_out.push_back(survive);
				if(survive==obin){continue;}
			}
			if(lit_flgs.setFlagNew(obin)){ LOG2("Append bin  " << obin.toDIMACS());
				clses.addBinPart(survive,obin);}
		}
		elim_hp	.removeIfContained(rep.toVar());
		wtching	.clear(rep); //will be updated later
		clses	.clearBinsPartOf(rep);
	}

	//if(lit_flgs.isSet(survive)){ units_out.push_back(survive);}

	Iter<Literal> it = clses.getBinsOf(survive); Variable v=INVALID_VAR;
	while(it.getNext(obin)){
		if(var_flgs.setFlagNew(v=obin.toVar())){update.push_back(v);}
		lit_flgs.carelessClear(obin);}
	lit_flgs.carelessClear(survive.flipped());
	asrt(lit_flgs.allCleared());
}

void SimplyFire::replAux(const Literal& lit, vector<Literal>& old_lits, vector<Literal>& new_lits){

	if(glob.isElim(lit)){
		new_lits.push_back(scc.getTmpLit(lit));
		LOG2("Repl lit " << lit.toDIMACS() << " in cls by " << new_lits.back().toDIMACS());

	}else{asrt(!lit_flgs.isSet(lit)); LOG2("keep lit " << lit.toDIMACS() << " in cls");
		lit_flgs.setFlag(lit);
		old_lits.push_back(lit);
	}
}


#define clean_ltr(y) {Variable x = y.toVar(); if(!glob.isElim(x) && var_flgs.setFlagNew(x)){udt_vars.push_back(x);}LOG2("Clean later " << x.toDIMACS());}
bool SimplyFire::replaceSCCs(SCC& scc, list<list<Literal> >& sccs, vector<Literal>& units_out){
	if(sccs.empty()) return true;

	sanity(asrtWatchingOK()); sanity(clses.asrtBinBothDirs(vals)); asrt(var_flgs.allCleared());
	sanity(asrtBinNoElims()); sanity(asrtOccCount());

	Literal survive; OccCls oc(ECP(0),false);
	vector<ECP> udt_cls; vector<Variable> udt_vars;

	//set scc vars replaced by its survivor
	for(list<list<Literal> >::iterator iter = sccs.begin();iter != sccs.end();){
		list<Literal>::iterator comp = iter->begin(); asrt(!glob.isElim(*comp));
		survive = *comp; LOG1("Replace vars surivor is lit " << survive.toDIMACS());
		if(vals.litKnwn(survive)){asrt(asrtKnownCount(*iter,iter->size()));
						iter = sccs.erase(iter); continue;}
		++comp;
		if(glob.isElim(*comp)){ asrt(asrtReplCount(*iter,iter->size()-1));
						iter = sccs.erase(iter); continue;}

		for(;comp != iter->end(); ++comp){
			Literal rep = *comp; LOG1("Replace lit "<< rep.toDIMACS() <<" surivor is " << survive.toDIMACS());
			if(rep == survive.flipped()){LOG1("Opposite lits in same scc");
				return glob.setInvalid("Opposite lits in same SCC");}

			scc.setTmpLit(rep,survive);
			scc.setTmpLit(rep.flipped(),survive.flipped());
			glob.setElim(rep.toVar());
			glob.elim_clses.remember(rep.flipped(),survive,VAR_ELIM);
			glob.elim_clses.remember(rep,survive.flipped(),VAR_ELIM);

			Iter<OccCls> oi = var_occs.getIter(rep.toVar());
			while(oi.getNext(oc)){
				ExtCls& ec = clses.getECls(oc.getECP());
				if(ec.isWaste() || ec.isMrkd()){ continue;}
				LOG2("Update cls later " << oc.getECP().idx);
				udt_cls.push_back(oc.getECP());	ec.setMrk(true);
			}
			var_occs.clear(rep.toVar());
			inc_stat(varrepl,1);
		}
		++iter;
	}

	//append binaries of replaced lits to the survivor of the component
	LOG2("Merge bins of replace to");
	while(!sccs.empty()){
		list<Literal>& comp = sccs.back(); asrt(comp.size()>1);
		//replace lits in one scc
		survive	= comp.front(); comp.pop_front();

		mergeBinLists(scc,survive		,comp,true,units_out,udt_vars);
		mergeBinLists(scc,survive.flipped()	,comp,false,units_out,udt_vars);

		sccs.pop_back();
	}


	//change replaced literals in binary lists
	LOG2("Update marked bins");
	for(uint32_t t=udt_vars.size();t--;){
		updateBins(scc,udt_vars[t].toNegLit(),units_out);
		updateBins(scc,udt_vars[t].toPosLit(),units_out);
		var_flgs.carelessClear(udt_vars[t]);
	}
	udt_vars.clear();
	LOG2("Finish update marked bins");

	sanity(clses.asrtBinBothDirs(vals)); sanity(asrtBinNoElims());


	//shrink and update clauses
	Literal wt[2]; Literal cnd; asrt(var_flgs.allCleared());
	LOG2("Shrink clauses"); vector<Literal> extra_lits; vector<Literal> add_lits;
	for(uint32_t c=0;c!=udt_cls.size();++c){
		ECP ecp 	= udt_cls[c]; 			LOG2("Handle cls " << ecp.idx);
		ExtCls& ec 	= clses.getECls(ecp); 	asrt(!ec.isWaste());
		ec			.setMrk(false);
		if(ec.isWaste()) continue;			asrt(extra_lits.empty()); asrt(add_lits.empty()); sanity(asrtPrintCls(ecp,"replace some ",ec.size()));

		Cls& cls	= glob.getCls(ec.getClp());

		const bool feli	= glob.isElim(ec.getLit());
		wt[feli]		= ec.getLit(); //don't allow that w[0] is replaced by w[1] -> this would cause a mess in watchers since both were assumed to be new watchers
		wt[!feli]		= cls.otherWatcher(wt[feli]); asrt(!glob.isElim(wt[0])||glob.isElim(wt[1]));
		LOG2("Watchers are " << wt[0].toDIMACS() << " and " << wt[1].toDIMACS());

		replAux(wt[0],add_lits,extra_lits);
		replAux(wt[1],add_lits,extra_lits);

		ClsLit* clit= cls.uwBeg(); LOG2("Replace uwtched ");
		do{
			replAux(Literal(clit->lit),add_lits,extra_lits);
		}while((clit++)->has_next);

		const uint32_t new_strt = add_lits.size();
		for(uint32_t k=0;k!=extra_lits.size();++k){
			cnd = extra_lits[k];
			if(lit_flgs.isSet(cnd.flipped())){ LOG2("New cls is taut ");
				ec.doWaste(); clean_ltr(wt[0]); clean_ltr(wt[1]);
				for(uint32_t t=add_lits.size();t--;){
					clean_ltr(add_lits[t]);
					lit_flgs.carelessClear(add_lits[t]);}

				goto NEXT_REPL_CLS;
			}

			if(lit_flgs.setFlagNew(cnd)){
				add_lits.push_back(cnd);
			}
		} asrt(!add_lits.empty());

		{const uint32_t nsz = add_lits.size(); asrt(nsz > 0);
		if(nsz<=2){
			lit_flgs.carelessClear(add_lits[0]);
			ec.doWaste(); clean_ltr(wt[0]); clean_ltr(wt[1]); clean_ltr(add_lits[0]);
			if(nsz==1){	LOG1("New clause is unit " << add_lits[0].toDIMACS());
				units_out.push_back(add_lits[0]); }
			else{		LOG2("New clause is bin " << add_lits[0].toDIMACS() << " and " << add_lits[1].toDIMACS());
				clses.addBin(add_lits[0],add_lits[1]); clean_ltr(add_lits[1]);
				lit_flgs.carelessClear(add_lits[1]);	}

		}else{
			const bool lrnt	= ec.isLearnt();
			signt_t sig		= 0;
			clit			= cls.uwBeg();		LOG2("Final clause is now with size " << nsz);

			for(uint32_t k=0;k!=nsz;++k){
				sig				|= ExtCls::getHash(cnd=add_lits[k]);
				lit_flgs		.carelessClear(cnd); LOG2(" " << cnd.toDIMACS());

				if(k>=new_strt)	{addOcc(cnd,ecp,lrnt);}

				if(k>1){		(clit++)->lit= cnd.idx();

				}else if(wt[k]!=cnd){
					wtching		.add(cnd,EWtch(ec.getClp(),ecp));
								LOG2("Take new wtcher " << cnd.toDIMACS() << " instead of " << wt[k].toDIMACS());
				}
			}

			asrt(clit>cls.uwBeg());
			(--clit)->has_next = 0;
			ec		.setSigSize(nsz,sig);
			ec		.setLit(add_lits[0]);
			cls		.setWatchers(add_lits[0],add_lits[1]);
			clses	.enqSubTest(ecp);
		}

		}NEXT_REPL_CLS:;
		extra_lits.clear(); add_lits.clear(); sanity(lit_flgs.allCleared());
	}

	//clean some lists
	LOG2("Clean some lists");
	for(uint32_t t=0;t!=udt_vars.size();++t){
		Variable v = udt_vars[t]; LOG2("Clean lists of " << v.toDIMACS());
		var_flgs.carelessClear(v);
		if(vals.varKnwn(v) || glob.isElim(v)) continue;
		cleanOccs(v);
		cleanWtches(v.toNegLit());
		cleanWtches(v.toPosLit());
	}

	asrt(var_flgs.allCleared()); sanity(asrtWatchingOK()); sanity(clses.asrtBinBothDirs(vals));
	asrt(clses.asrtNoMrkdClses()); sanity(asrtOccCount());
	return true;
}

/******************************************
 * Subsumption checks
 *******************************************/


bool SimplyFire::elimAndSubsum(const bool blkd_chk){
	uint32_t cnt = 0; uint32_t check;

	do{
		if(!subsumCheck(cnt)) {	return glob.setInvalid("Bad subsum check");}
		check = cnt;

		if(!elim(blkd_chk, cnt)){return glob.setInvalid("Bad var elim");}
		check += cnt;
	}while(check);

	return true;
}


bool SimplyFire::subsumCheck(uint32_t& cnt){
	Literal lit; ECP ecp; uint32_t lst; cnt = 0;
	T_ON(subchk);
	do{ asrt(units_unpropd.empty());
		lst = cnt;
		while(clses.dequeueSubLit(lit)){
			if(vals.litKnwn(lit)) continue;

			collectBinSubsumed(lit); ++cnt;
			if(!propagateUnpropdUnits()){ T_OFF(subchk);
				return glob.setInvalid("Bad Unit after collectBinSubsumed");}
		}

		if(clses.dequeueSubCls(ecp)){
			collectSubsumed(ecp); ++cnt;} asrt(units_unpropd.empty());

#ifdef PRINT_ELIM_SUB
		if(glob.sets.print_console){
			printf("Bins to check %d and clauses to check: %d \r",clses.binsQSize(),clses.subQSize());}
#endif

		cnt += delClses();
	}while(lst != cnt);
	T_OFF(subchk);
	return true;
}

#define check_lit(x) \
	if(lit_flgs.isSet(lit)){ if(--sub_marked==0){ return flp;}}	\
	else if(lit_flgs.isSet(lit.flipped())){ \
		if(flp != TRUE_LIT){	return FALSE_LIT;} \
		if(--sub_marked==0){ 	return lit;} \
		flp = lit;}

/**
 * Checks if clause "sub" is a subset of clause "super"
 * Literals of "sub" are assumed to be marked
 * Returns TRUE_LIT if "sub" subsumes "super"
 * FALSE_LIT if there is no subset relation or any Literal "a \in super" that can be removed from
 * "super" by backward subsumption
 */
Literal SimplyFire::chkSuperOrEqToMarked(const uint32_t& sub_sig, uint32_t sub_marked, ExtCls& sup){
	asrt(sub_marked > 1); asrt(!sup.isMrkd()); asrt(!sup.isWaste()); asrt(sup.size()>2);

	if(sup.cannotBeSubsumedBy(sub_marked,sub_sig)){	return FALSE_LIT;}

	Literal flp 	= TRUE_LIT;
	Literal lit		= sup.getLit();
	check_lit(lit);
	Cls& cls		= glob.getCls(sup.getClp());
	lit			= cls.otherWatcher(lit);
	check_lit(lit);

	ClsLit* clit 	= cls.uwBeg();
	do{
		lit = Literal(clit->lit);
		check_lit(lit);
	}while((clit++)->has_next);

	asrt(sub_marked);
	return FALSE_LIT;
}

#define bin_chk(x) if(lit_flgs.isSet(x)){ LOG2(x.toDIMACS() << " is set return sub"); return TRUE_LIT;} \
		if(allow_flpd && lit_flgs.isSet(x.flipped())){ LOG2("flipped of " << x.toDIMACS() << " is set");  return x;}
/*
 * assumes that for binary clauses (sub_bin,x) all x are marked
 * if super_cand is subsumed TRUE_LIT is returned, if nothing is subsumed then FALSE_LIT is returned
 * if one lit of super_cand can be removed by backward subsumption this lit is returned
 */
Literal SimplyFire::chkSuperToBins(const Literal& sub_bin, const signt_t& bins_msk, ExtCls& sup,const bool allow_flpd){
	asrt(!lit_flgs.isSet(sub_bin)); asrt(!lit_flgs.isSet(sub_bin.flipped())); asrt(!sup.isWaste());  asrt(!sup.isMrkd());
	LOG2("Check bins of " << sub_bin.toDIMACS() << " allow flip " << allow_flpd);

	if(!sup.mayContainOneOf(bins_msk)){	return FALSE_LIT;}

	Literal lit		= sup.getLit();
	bin_chk(lit);
	Cls& cls		= glob.getCls(sup.getClp());
	lit			= cls.otherWatcher(lit);
	bin_chk(lit);

	ClsLit* clit 	= cls.uwBeg();
	do{
		lit = Literal(clit->lit);
		bin_chk(lit);
	}while((clit++)->has_next);

	return FALSE_LIT;
}

void	   SimplyFire::removeVarInAux(const Variable& v, vector<ECP>& v_cls){
	LOG2("remove var in all clses helper:" << v.toDIMACS());
	if(v_cls.empty()) {sanity(asrtOccCount()); return;}

	uint32_t s =0; sanity(asrtOccCount(v.idx()));
	for(uint32_t i=0;i!=v_cls.size();++i){
		ExtCls& ec	= clses.getECls(v_cls[i]);
		if(ec.doWasteWasAlve()){		LOG2("Tmp del cls " << v_cls[i].idx << " with size " << ec.size());
			v_cls[s++]=v_cls[i];}else{	LOG2("Cls already wst " << v_cls[i].idx << " with size " << ec.size());}
	}
	cleanOccs(v);
	while(s--){
		ExtCls& ec = clses.getECls(v_cls[s]); LOG2("wake up tmp " << v_cls[s].idx);
		ec.undoWaste(); ec.setMrk(false);}
	sanity(asrtOccCount()); sanity(clses.asrtNoMrkdClses());
}

void	  SimplyFire::collectBinSubsumed(const Literal& lit){
	asrt(vals.litUkwn(lit)); sanity(lit_flgs.allCleared());sanity(asrtOccCount()); sanity(clses.asrtNoMrkdClses());
	LOG2("Collect bin subsumed start with " << lit.toDIMACS());

	vector<Literal> clean; signt_t sig = 0; Literal next = lit; uint32_t q = 0;
	//mark all binary partners of lit
	while(true){
		Iter<Literal> it = clses.getBinsOf(next);
		while(it.getNext(next)){
			if(lit_flgs.setFlagNew(next)){	clean.push_back(next); sig |= ExtCls::getHash(next);}
		}
		if(q>=clean.size()) break;
		next = clean[q++].flipped();
	}
	if(lit_flgs.isSet(lit)){LOG1("Unit found while collectBinSubsumed: " << lit.toDIMACS());
		units_unpropd.push_back(lit);
		goto CLEAN_BINS;}

	lit_flgs.clear(lit.flipped()); //just delete tautological clauses

	{vector<ECP> shrnkd; const Variable thisv(lit.toVar());
	Iter<OccCls> iter = var_occs.getIter(thisv); OccCls oc(ECP(0),false); const bool pol = lit.isPos();
	while(iter.getNext(oc)){
		ECP	supp			= oc.getECP();
		ExtCls& sup			= clses.getECls(supp);
		if(sup.isWaste() || sup.isMrkd()) continue;

		const bool right_pol= oc.isPosOcc() == pol;
		Literal rem			= chkSuperToBins(lit,sig,sup,right_pol);
		if(rem==FALSE_LIT)	continue; // no subsumption

		if(rem==TRUE_LIT){
			if(right_pol){inc_stat(sub_work,1);	for_deletion.push_back(Subsumd(supp,ECP(max32))); continue;}
			rem = lit.flipped(); LOG2("Flip lit to remove in Bin subsum check");}

		inc_stat(bwsub_litselim,1); sanity(asrtOccCount(thisv.idx()));aux_var(uint32_t sup_sz = )
		removeLit(supp,rem); asrt(sup_sz>1); //empty clause can not be learnt here

		if(rem.toVar()==thisv){
			sup.setMrk(true);
			shrnkd.push_back(supp);} //avoid concurrent modification problems
		else{
			bool revive	= sup.doWasteWasAlve();
			cleanOccs(rem.toVar());
			if(revive){sup.undoWaste();} asrt(asrtOccCountOf(rem.toVar()));
		}
		sanity(asrtOccCount(thisv.idx()));//check for all but lit
	}

	//clean occurrence list of thisv
	removeVarInAux(thisv,shrnkd);} sanity(asrtOccCount());

	CLEAN_BINS:;
	for(uint32_t q=clean.size();q--;){lit_flgs.carelessClear(clean[q]);} asrt(lit_flgs.allCleared());
}




void	SimplyFire::collectSubsumed(const ECP& subp){
	sanity(lit_flgs.allCleared()); sanity(asrtOccCount()); sanity(clses.asrtNoMrkdClses());
	ExtCls& subc	= clses.getECls(subp);
	if(subc.isWaste())	return;
	LOG2("Collect subsumed of cls " << subp.idx << " with size " << subc.size()); inc_stat(subchk,1);


	uint32_t sub_sz	= subc.size(); asrt(sub_sz>2);
	signt_t sub_sig	= subc.getSig();

	Literal least 		= subc.getLit(); lit_flgs.setFlag(least);
	uint32_t min 		= elim_hp.getValue(least.toVar()).sum();
	Cls& cls			= glob.getCls(subc.getClp());
	Literal lit			= cls.otherWatcher(least); lit_flgs.setFlag(lit);
	uint32_t occ_sz		= elim_hp.getValue(lit.toVar()).sum();
	if(occ_sz<min)		{min=occ_sz;least=lit;}
	ClsLit* clit		= cls.uwBeg();

	//find variable with least occurrences
	do{
		lit 			= Literal(clit->lit); lit_flgs.setFlag(lit);
		occ_sz			= elim_hp.getValue(lit.toVar()).sum();
		if(occ_sz<min)	{min=occ_sz;least=lit;}
	}while((clit++)->has_next);

	vector<ECP> shrnkd; Variable thisv = least.toVar();

	Iter<OccCls> iter = var_occs.getIter(thisv); OccCls oc(subp,false);
	while(iter.getNext(oc)){
		ECP	supp		= oc.getECP();
		if(supp==subp) 		continue;
		ExtCls& sup		= clses.getECls(supp);
		if(sup.isWaste() || sup.isMrkd()) continue;

		Literal rem			= chkSuperOrEqToMarked(sub_sig,sub_sz,sup);

		if(rem==FALSE_LIT)	continue; // no subsumption

		if(rem==TRUE_LIT){	inc_stat(sub_work,1);
			for_deletion.push_back(Subsumd(supp,subp)); continue;}

		inc_stat(bwsub_litselim,1);
		uint32_t sup_sz 		= removeLit(supp,rem); asrt(sup_sz>1); //empty clause can not be learnt here

		if(rem.toVar()==thisv){
			sup.setMrk(true);
			shrnkd.push_back(supp);} //avoid concurrent modification problems
		else{
			bool revive	= sup.doWasteWasAlve();
			cleanOccs(rem.toVar());
			if(revive){	sup.undoWaste();} asrt(asrtOccCountOf(rem.toVar()));
		}

		if(sup_sz<sub_sz){ 	//now sub is subsumed by the shrinked super clause
			inc_stat(bwsub_self,1);
			for_deletion.push_back(Subsumd(subp,supp));
			break;
		}
		sanity(asrtOccCount(thisv.idx()));//check for all but thisv
	}

	lit_flgs.carelessClear(lit=subc.getLit());
	lit_flgs.carelessClear(cls.otherWatcher(lit));
	clit = cls.uwBeg();
	do{ lit_flgs.carelessClear(Literal(clit->lit));}while((clit++)->has_next);

	//clean occurrence list of thisv
	removeVarInAux(thisv,shrnkd);
	sanity(asrtOccCount());
}

//returns the size of the new clause
uint32_t	SimplyFire::removeLit(ECP& ecp, const Literal& lit){
	ExtCls& ec	= clses.getECls(ecp); asrt(ec.size()>2); asrt(!ec.isWaste());
	LOG2("Remove lit " << lit.toDIMACS() << " in cls " << ecp.idx << " with size " << ec.size());

	Cls& cls		= glob.getCls(ec.getClp());
	Literal 		wt[2];
	wt[0]		= ec.getLit();
	wt[1]		= cls.otherWatcher(wt[0]);
	ClsLit* clit	= cls.uwBeg();
	if(wt[0]==lit || wt[1]==lit){asrt(!ec.isWaste()); LOG2("Shall remove watcher " << lit.toDIMACS());
		ec.doWaste();
		cleanWtches(lit);
		ec.undoWaste();
		bool idx			= (wt[1]==lit);
		wt[idx] 			= Literal(clit->lit); asrt(wt[0]!=lit); asrt(wt[1]!=lit); asrt(wt[0]!=wt[1]);
		clit->lit 			= lit.idx();
		cls				.setWatchers(wt[0],wt[1]);
		wtching			.add(wt[idx],EWtch(ec.getClp(),ecp));
		ec				.setLit(wt[0]);
	}
	LOG2("Keep lit " << wt[0].toDIMACS());  LOG2("Keep lit " << wt[1].toDIMACS());

	signt_t sig 	= ExtCls::getHash(wt[0]) | ExtCls::getHash(wt[1]);
	ClsLit* cp = clit;
	do{
		Literal l(clit->lit);
		if(l!=lit){
			sig 		|= ExtCls::getHash(l);
			(cp++)->lit	= clit->lit;  LOG2("Keep lit " << l.toDIMACS());
		}else{LOG2("Remove lit " << l.toDIMACS());}	//if(fix){		decElimHp(l);} //will be reset in cleanOccs

	}while((clit++)->has_next);

	const uint32_t sz = (cp - cls.uwBeg()) +2 ;
	LOG2("Remove of " << lit.toDIMACS() << " makes new size " << sz); asrt(sz>1); asrt(ec.size()==sz+1);

	if(sz==2){  asrt(wt[0] != lit); asrt(wt[1] != lit); LOG2("Cls became binary " << wt[0].toDIMACS() << " , " << wt[1].toDIMACS());

		if(!ec.isLearnt()){decElimHp(wt[0]); decElimHp(wt[1]);	}
		clses.addBin(wt[0],wt[1]);
		ec.doWaste();
		cleanWtches(wt[0]);
		cleanWtches(wt[1]);
	}else{
		(--cp)->has_next = 0;
		ec.setSigSize(sz,sig);
		clses.enqSubTest(ecp);
	}

	sanity(asrtWatchingOK());
	return sz;
}

/***********************************************
 * Handle proposal of unit clauses and subsumed clauses
 ***********************************************/
bool SimplyFire::propagate(Literal tlit,vector<Literal>& clean){
	if(vals.litKnwn(tlit)){LOG2("Found fix to propagate " << tlit.toDIMACS());
		return vals.litTrue(tlit);}

	Literal ol,cnd; EWtch ew(ClsP(0),ECP(0)); asrt(uq_begin<=units_out.size());  sanity(asrtWatchingOK());

	enqueue(tlit);
	while(uq_begin < units_out.size()){
		tlit 				= units_out[uq_begin++];LOG2("Make lit fix " << tlit.toDIMACS());
		const Literal flit	= tlit.flipped();

		Iter<Literal> bi	= clses.getBinsOf(flit);
		while(bi.getNext(ol)){
			if(vals.litUkwn(ol)){
				enqueue(ol);
			}else if(vals.litFalse(ol)){
				LOG1("Prop of " << tlit.toDIMACS() << " problems with binary " << flit.toDIMACS() << " , " << ol.toDIMACS());
				return glob.setInvalid("Binary clause causes conflict in simplification");}
		}

		WIter<EWtch> ci		= wtching.getWIter(flit);
		while(ci.getNext(ew)){
			asrt(!clses.getECls(ew.extp).isWaste());

			Cls& cls	= glob.getCls(ew.clp);
			ol			= cls.otherWatcher(flit); asrt(flit!=ol);
			LOG2("Check cls at " << ew.clp.idx() << " when prop lit " << tlit.toDIMACS());

			if(vals.litTrue(ol)){LOG2("Keep cls other true: " << ol.toDIMACS());
					ci.keep(ew);  continue;}

			ClsLit* clit	= cls.uwBeg();
			do{
				Literal cnd(clit->lit);
				if(!vals.litFalse(cnd)){
					LOG2("Found new watch " << cnd.toDIMACS());
					cls.changeWatchers(ol,clit,flit); asrt(cnd!=ol);
					wtching.add(cnd,ew);
					clses.setLitFor(ew.extp,cnd);
					goto NEXT_WTCHD;
				}
			}while((clit++)->has_next);

			if(vals.litFalse(ol)){return glob.setInvalid("Empty clause in simplification");}

			LOG2("Enqueue other: " << ol.toDIMACS());
			ci.keep(ew);
			enqueue(ol);

			NEXT_WTCHD:;
		}

		wtching.finish(flit,ci);

		addBinsSetFlgs(tlit,clean);
		clses	.clearBinsPartOf(tlit);
		clses	.clearBinsPartOf(flit);
	}
	return true;
}

#define to_clean(x) if((lit=x,lit_flgs.setFlagNew(lit))){clean.push_back(lit);}
bool	SimplyFire::propagateUnpropdUnits(){
	asrt(lit_flgs.allCleared()); sanity(asrtWatchingOK()); sanity(asrtOccCount());

	if(units_unpropd.empty()) return true;
	//propagate
	vector<Literal> clean; uint32_t strt = units_out.size();
	for(uint32_t i=0;i!=units_unpropd.size();++i){
		if(!propagate(units_unpropd[i],clean)) return false;
	}
	units_unpropd.clear();

	//clean all sat binaries
	cleanBinsAndFlgs(clean);

	vector<ECP> shrink; OccCls oc(ECP(0),0); Literal tlit,lit; asrt(lit_flgs.allCleared());
	//waste satisfied clauses
	while(strt!=units_out.size()){
		tlit = units_out[strt++];
		const bool pos = tlit.isPos();

		Iter<OccCls> oi = var_occs.getIter(tlit.toVar());
		while(oi.getNext(oc)){
			ExtCls& ec = clses.getECls(oc.getECP());
			if(ec.isWaste()) continue;

			if(oc.isPosOcc()==pos){
				Cls& cls		= glob.getCls(ec.getClp());
				ec				.doWaste();
				Literal w2 		= ec.getLit();
				to_clean(w2); 	to_clean(cls.otherWatcher(w2));
				ClsLit* clit 	= cls.uwBeg();
				do{to_clean(Literal(clit->lit));}while((clit++)->has_next);

			}else if(!ec.isMrkd()){ //shrink cls in next step
				ec.setMrk(true);
				shrink.push_back(oc.getECP());
			}
		}
		var_occs	.clear(tlit.toVar());
		wtching	.clear(tlit);
		wtching	.clear(tlit.flipped());
	}

	//shrink clauses with false lit
	signt_t sig; bool shrnkd;
	for(uint32_t i=0;i!=shrink.size();++i){
		ExtCls& ec 	= clses.getECls(shrink[i]);
		ec			.setMrk(false);
		if(ec.isWaste()) continue;

		LOG2("Shrink clause at " << ec.getClp().idx());
		Cls& cls	= glob.getCls(ec.getClp());
		Literal w2 	= ec.getLit();
		Literal w1	= cls.otherWatcher(w2);
		const int32_t uw_cnt = removeFalseLits(cls,w1,w2,sig,shrnkd); asrt(shrnkd);
		if(uw_cnt <=0){
			if(uw_cnt==0){LOG2("Binary Propagate shrink"); clses.addBin(w1,w2);}
			ec.doWaste();
			to_clean(w1); to_clean(w2);
		}else{
			ec.setSigSize(uw_cnt+2, sig);
			clses.enqSubTest(shrink[i]);
		}
	}

	//clean others occurrence and watched lists
	for(uint32_t i=0;i!=clean.size();++i){
		Literal lit = clean[i];
		lit_flgs	.clear(lit);
		if(vals.litKnwn(lit)) continue;
		cleanWtches(lit);
		if(!lit_flgs.isSet(lit.flipped())){
			cleanOccs(lit.toVar());}
	}

	asrt(lit_flgs.allCleared()); sanity(asrtWatchingOK()); sanity(asrtOccCount());
	return true;
}


bool	SimplyFire::delClses(){
	asrt(lit_flgs.allCleared()); sanity(asrtWatchingOK()); sanity(asrtOccCount());
	LOG2("delete subsumed and other clauses. Total: " << for_deletion.size());

	if(for_deletion.empty()) return false;
	vector<Literal> clean;

	for(uint32_t i=0;i!=for_deletion.size();++i){
		Subsumd& s 	= for_deletion[i];
		ExtCls& del	= clses.getECls(s.who);
		if(del.isWaste()) continue;

		LOG2("cls " << s.who.idx << " is subsumed by " << s.by_whom.idx); sanity(asrtPrintCls(s.who,"cls to die ", s.who.idx));


		Literal w1 	= del.getLit();
		Cls& cls	= glob.getCls(del.getClp());
		Literal w2 	= cls.otherWatcher(w1);
		if(lit_flgs.setFlagNew(w1)){clean.push_back(w1);}
		if(lit_flgs.setFlagNew(w2)){clean.push_back(w2);}

		if(!del.isLearnt()){
			LOG2("Cls " << s.who.idx << " is fix");
			ensureFix(s.by_whom); sanity(asrtOccCount());

			decElimHp(w1); decElimHp(w2);
			ClsLit* clit = cls.uwBeg();
			do{decElimHp(Literal(clit->lit));}while((clit++)->has_next);

		}
		del.doWaste();
		sanity(asrtOccCount());
	}

	for(uint32_t i=clean.size();i--;){
		lit_flgs.carelessClear(clean[i]);
		cleanWtches(clean[i]);
	}

	for_deletion.clear();
	asrt(lit_flgs.allCleared()); sanity(asrtWatchingOK());sanity(asrtOccCount());
	return true;
}


/***********************************************
 * Basics on Data Structure
 ***********************************************/

#define add_if(x){if(x.toVar() != ignore){out.push_back(x);}}
void SimplyFire::addAllLitsBut(const ECP& ecp, const Variable& ignore, vector<Literal>& out){
	ExtCls& ec 	= clses.getECls(ecp);
	Cls& cls	= glob.getCls(ec.getClp());
	Literal lit	= ec.getLit();
	add_if(lit);
	lit			= cls.otherWatcher(lit);
	add_if(lit);

	ClsLit* clit = cls.uwBeg();
	do{
		Literal l(clit->lit);
		add_if(l);
	}while((clit++)->has_next);
}


void SimplyFire::addBinsSetFlgs(const Literal& lit, vector<Literal>& out){
	Literal ol;
	Iter<Literal> bit = clses.getBinsOf(lit);
	while(bit.getNext(ol)){
		if(lit_flgs.setFlagNew(ol)){out.push_back(ol);}
	}
}

void SimplyFire::cleanBinsAndFlgs(vector<Literal>& clean){
	Literal lit,ol; LOG2("Clean bins");

	for(uint32_t t=clean.size();t--;){
		lit_flgs.carelessClear(lit=clean[t]);
		if(vals.litKnwn(lit)){clses.clearBinsPartOf(lit);continue;}

		WIter<Literal> bi = clses.getWBinsOf(lit);
		while(bi.getNext(ol)){
			if(vals.litUkwn(ol) && !glob.isElim(ol)){bi.keep(ol);}
			else{LOG2("Binary dies in " << lit.toDIMACS() << " who " << ol.toDIMACS());}
		}
		clses.finishBins(lit,bi);
	}
	clean.clear(); sanity(lit_flgs.allCleared());
}


int32_t  SimplyFire::removeFalseLits(Cls& cls, const Literal& w1, const Literal& w2, signt_t& sig,bool& shrinkd){

	if(vals.litKnwn(w1) || vals.litKnwn(w2)){asrt(vals.litTrue(w1)|| vals.litTrue(w2));
		return -1;}

	ClsLit* clit = cls.uwBeg(); ClsLit* cp = clit;
	sig			= ExtCls::getHash(w1) | ExtCls::getHash(w2);
	do{
		if(vals.litTrue(Literal(clit->lit))){ return -1;}
		if(!vals.litFalse(Literal(clit->lit))){
			sig |= ExtCls::getHash(Literal(clit->lit));
			(cp++)->lit = clit->lit;}
	}while((clit++)->has_next);

	const uint32_t uw_cnt = cp - cls.uwBeg();
	shrinkd				  = cp != clit;

	//new cls is binary - delete this clause and keep binary
	if(uw_cnt>0){(--cp)	->has_next = 0;}
	return uw_cnt;
}

void	SimplyFire::cleanOccs(const Variable& var){
	LOG2("Clean occs of " << var.toDIMACS());
	OccCls oc(ECP(0),0); asrt(!vals.varKnwn(var));
	WIter<OccCls> oi = var_occs.getWIter(var);
	OccCount cnt;
	while(oi.getNext(oc)){
		ExtCls& ec = clses.getECls(oc.getECP());
		if(ec.isWaste()){LOG2("Remove cls " << oc.getECP().idx << " of var " << var.toDIMACS());
			continue;}
		oi.keep(oc);
		if(!ec.isLearnt()){
			++cnt.getOcc(oc.isPosOcc());}
	}
	var_occs.finish(var,oi);
	if(cnt != elim_hp.getValue(var)){
		elim_hp		.setValue(var,cnt);
		elim_hp		.restoreElementIf(var);}
}

void	SimplyFire::cleanWtches(const Literal& lit){
	LOG2("Clean watchers of " << lit.toDIMACS());
	EWtch ewtch(ClsP(0),ECP(0)); asrt(!vals.litKnwn(lit));
	WIter<EWtch> ew = wtching.getWIter(lit);
	while(ew.getNext(ewtch)){
		ExtCls& ecl 	= clses.getECls(ewtch.extp);
		if(!ecl.isWaste()){ew.keep(ewtch);	}
	}
	wtching.finish(lit,ew);
}


void	SimplyFire::addOcc(const Literal& to, const ECP& ecp, const bool lrnt){
	LOG2("Add occur " << ecp.idx << " to " << to.toDIMACS() << (lrnt?" lrnt":" fix"));
	var_occs.add(to.toVar(),OccCls(ecp,to.isPos()));
	if(!lrnt){incElimHp(to);} sanity(asrtOccsContain(to,ecp));
}

void	SimplyFire::ensureFix(ExtCls& ec){
	if(ec.isWaste() || !ec.isLearnt()) return;
	LOG2("Make learnt cls fix ");
	Literal lit		= ec.getLit();
	incElimHp(lit);

	Cls& cls		= glob.getCls(ec.getClp());
	lit				= cls.otherWatcher(lit);
	incElimHp(lit);
	ClsLit* clit 	= cls.uwBeg();
	do{incElimHp(Literal(clit->lit));}while((clit++)->has_next);
	ec.setFix();
	cls.setLrnt(false);
}

ECP		SimplyFire::newCls(Literal lits[], const uint32_t sz, bool lrnt, const uint32_t lbd, bool enq, bool assym){
	asrt(sz > 2);
	LOG2("Create new cls with size " << sz);
	signt_t sig = 0; for(uint32_t i=sz;i--;){sig |= ExtCls::getHash(lits[i]);}
	ClsP clp 	= glob.clsCreate(lits,sz,false,0);
	ECP ecp 	= clses.addCls(clp,lits[0],sz,lrnt,sig,enq,false,assym);
	registerOcc(lits[0],lits[1],ecp,glob.getCls(clp),false);

	wtching.add(lits[0],EWtch(clp,ecp));
	wtching.add(lits[1],EWtch(clp,ecp)); sanity(asrtPrintCls(ecp,"brand new cls with size ",sz));
	return ecp;
}


void	SimplyFire::registerOcc(const Literal& w1, const Literal& w2, const ECP& ecp, Cls& cls,const bool lrnt){
	addOcc(w1,ecp,lrnt);
	addOcc(w2,ecp,lrnt);
	ClsLit* clit = cls.uwBeg();
	do{addOcc(Literal(clit->lit),ecp,lrnt);}while((clit++)->has_next);
}


void SimplyFire::incElimHp(const Literal& l){
	Variable var = l.toVar();
	if(vals.varKnwn(var) || glob.isElim(var)) return;
	OccCount cnt = elim_hp.getValue(var);
	LOG2("Increment literal "<<l.toDIMACS() << " neg cnt is " << cnt.getOcc(0) << " pos count is " << cnt.getOcc(1));
	++cnt.getOcc(l.isPos());
	elim_hp.setValue(var,cnt);
	elim_hp.restoreElementIf(var);
	LOG2("Incremented now neg cnt is " << cnt.getOcc(0) << " pos count is " << cnt.getOcc(1));
	//if(cnt.getOcc(false)>MAX_ELIM_FACTOR && cnt.getOcc(true)>MAX_ELIM_FACTOR ){
	//	elim_hp.removeIfContained(var);}
}

void SimplyFire::decElimHp(const Literal& l){
	Variable var = l.toVar();
	if(vals.varKnwn(var) || glob.isElim(var)) return;
	LOG2("Decrement variable of lit "<<l.toDIMACS() << " with pos: " << elim_hp.getValue(var).getOcc(true) << " and neg: " << elim_hp.getValue(var).getOcc(false));

	OccCount cnt = elim_hp.getValue(var); asrt(cnt.getOcc(l.isPos())>0);
	--cnt.getOcc(l.isPos());
	elim_hp	.setValue(var,cnt);
	LOG2("Decremented value of " << l.toDIMACS() << " now pos: " << elim_hp.getValue(var).getOcc(true) << " and neg: " << elim_hp.getValue(var).getOcc(false));
	elim_hp	.restoreElementIf(var); //reinsert since things might have changed for var elimination and blocked check
}







/*******************************
 * Assertions and Sanity
 *******************************/



bool SimplyFire::asrtWatchingOK()const{
	bool ok = true; EWtch cand(ClsP(0),ECP(0));
	for(uint32_t i=wtching.listCount();i--;){
		Literal lit(i); bool fls = vals.litFalse(lit);
		if(glob.isElim(lit) || vals.litKnwn(lit)) continue;
		Iter<EWtch> it 	= wtching.getIter(lit);
		while(it.getNext(cand)){

			if(clses.getECls(cand.extp).isWaste()){
				ok = false;
				LOG0("Cls " << cand.extp.idx << " is already wasted and listed in " << lit.toDIMACS());
				continue;}

			Cls& cls  	= glob.getCls(cand.clp);
			Literal ol	= cls.otherWatcher(lit);
			if(ol==lit){
				LOG0("Cls " << cand.clp.idx() << " bad watching. Seen from " << lit.toDIMACS());
				ok = false; continue;}

			if(fls && !vals.litTrue(ol)){ok=false;
				LOG0("Cls " << cand.clp.idx() << " has false wtcher " << lit.toDIMACS() <<" and " <<
							(vals.litFalse(ol)?"false":"ukwn") << " lit " <<ol.toDIMACS());}

			uint32_t cnt = asrtWatchingCount(ol,cand);

			if(cnt==0){ ok=false;
				LOG0("Cls " << cand.clp.idx() << " only watched by " << lit.toDIMACS() << " not by " << ol.toDIMACS());
			}else if(cnt!=1){ ok=false;
				LOG0("Cls " << cand.clp.idx() << " watched by " << ol.toDIMACS() << ": " << cnt << " times. Seen from " << lit.toDIMACS());
			}

			ClsLit* clit = cls.uwBeg();
			do{
				Literal l(clit->lit);
				cnt = asrtWatchingCount(l,cand);
				if(cnt != 0){ok=false;
					LOG0("Cls " << cand.clp.idx() << " watched by nonwatcher " << l.toDIMACS() << ": "<< cnt << " times. Seen from " << lit.toDIMACS());
				}
			}while((clit++)->has_next);
		}
	}

	LOG1("SimplyFire::asrtWatchingOK is " << (ok?"ok":"bad"));
	return ok;
}

uint32_t SimplyFire::asrtWatchingCount(const Literal& a, const EWtch& ew)const{
	uint32_t c=0;
	Iter<EWtch> it = wtching.getIter(a); EWtch cand(ClsP(0),ECP(0));
	while(it.getNext(cand)){c += (cand==ew);}
	return c;
}

bool SimplyFire::asrtKnownCount(const list<Literal>& scc, const uint32_t cnt){
	uint32_t c = 0;
	for(list<Literal>::const_iterator it = scc.begin();it != scc.end();++it){
		c += vals.litKnwn(*it);
	}
	return c == cnt;
}

bool SimplyFire::asrtReplCount(const list<Literal>& scc, const uint32_t cnt){
	uint32_t c = 0;
	for(list<Literal>::const_iterator  it = scc.begin();it != scc.end();++it){
		c += glob.isElim(it->toVar());
	}
	return c == cnt;
}

bool	SimplyFire::asrtBinNoElims() const {
	bool ok = true; Literal cand;
	for(uint32_t i=lit_flgs.size();i--;){
		Literal lit(i);
		if(vals.litKnwn(lit)) continue;

		Iter<Literal> it = clses.getBinsOf(lit);
		while(it.getNext(cand)){
			if(vals.litKnwn(cand)){
				LOG0("Contains bin with known lit: " << lit.toDIMACS() << " to " << cand.toDIMACS());
				ok=false; continue;
			}
			if(glob.isElim(cand)){
				LOG0("Contains bin with eliminated lit: " << lit.toDIMACS()  << " to " << cand.toDIMACS());
				if(glob.isElim(lit)){LOG0("Lit " << lit.toDIMACS() << " is also eliminated");	}
				else{LOG0("Lit " << lit.toDIMACS() << " is not eliminated");	}
				ok=false; continue;
			}
		}
	}
	LOG1("ExtClses::asrtBinBothDirs is " << (ok?"ok":"bad"));

	return ok;
}

void SimplyFire::printBinaries()const {
	Literal cand;
	for(uint32_t i=lit_flgs.size();i--;){
		Literal lit(i);
		cout << "c [" << lit.toDIMACS() << "]";
		char del = ':';

		Iter<Literal> it = clses.getBinsOf(lit);
		while(it.getNext(cand)){
			cout << del << " " << cand.toDIMACS() ;
			if(vals.litKnwn(cand)){cout << " <fix>";}
			if(glob.isElim(cand)){ cout << " <elim>";}
			del = ',';
		}
		cout << endl;
	}
}


bool SimplyFire::asrtOccCountOf(const Variable& var) {
	if(vals.varKnwn(var) || glob.isElim(var)) return true;
	bool ok = true;
	OccCls oc(ECP(0),0);
	Iter<OccCls> oi = var_occs.getIter(var); OccCount cnt; uint32_t lrnt = 0;
	while(oi.getNext(oc)){
		ExtCls& ec = clses.getECls(oc.getECP());
		if(ec.isWaste()){continue;}
		if(ec.isLearnt()){++lrnt;}
		else{++cnt.getOcc(oc.isPosOcc());}

		ok = asrtClsContains(oc.getECP(),ec,oc.isPosOcc()? var.toPosLit() : var.toNegLit()) && ok;
	}
	OccCount stored = elim_hp.getValue(var);
	if(cnt.getOcc(true) != stored.getOcc(true) || cnt.getOcc(false) != stored.getOcc(false) ){
		LOG0("Bad occ count for variable " << var.toDIMACS() << " with " << lrnt << " learnt clauses:");
		LOG0(	"Pos occ is " << cnt.getOcc(true) << " stored is " << stored.getOcc(true) <<
				" Neg occ is " << cnt.getOcc(false) << " stored is " << stored.getOcc(false));
		return false;
	}
	if(!ok){
		LOG0("Not all listed clauses do really contain the var");
		return false;}

	return true;
}

bool SimplyFire::asrtClsContains(ECP ecp, ExtCls& ec,const Literal& who){
	if(ec.isWaste()){LOG0("Cls is waste and cannot contain " << who.toDIMACS());return false;}
	Literal lit = ec.getLit();
	if(lit==who) return true;
	Cls& cls	= glob.getCls(ec.getClp());
	if(cls.otherWatcher(lit)==who) return true;

	ClsLit* clit = cls.uwBeg();
	do{
		if(Literal(clit->lit)==who) return true;
	}while((clit++)->has_next);

	LOG0("Cls "<< ecp.idx<< " does not contain " << who.toDIMACS());
	return false;
}

bool SimplyFire::asrtOccCount(const uint32_t& ignore){
	bool ok = true;
	for(uint32_t i=0;i!=var_occs.listCount();++i){
		if(i==ignore) continue;
		ok = asrtOccCountOf(Variable(i)) && ok;
	}
	if(ok){LOG2("occ counts all valid with ignore " << ignore);}
	return asrtOccsContainAll() && ok;
}


bool 	SimplyFire::asrtPrintCls(const ECP& e, string what, int val){
	LOG2(what <<  val << " cls: ");
	ExtCls& ec 	= clses.getECls(e);
	Literal lit	= ec.getLit();
	LOG2(lit.toDIMACS() << "");

	Cls& cls	= glob.getCls(ec.getClp());
	lit			= cls.otherWatcher(lit);
	LOG2(lit.toDIMACS() << "");

	ClsLit* clit= cls.uwBeg();
	do{
		lit = Literal(clit->lit);
		LOG2(lit.toDIMACS() << "");
	}while((clit++)->has_next);
	return true;
}

bool SimplyFire::asrtOccsContain(const Literal& l, const ECP& e){
	if(vals.litKnwn(l) || glob.isElim(l)) return true;

	Iter<OccCls> it = var_occs.getIter(l.toVar());
	OccCls cnd(ECP(0),false);
	uint32_t cnt = 0; uint32_t cntf =0;
	while(it.getNext(cnd)){
		if(cnd.getECP() == e){
			if(cnd.isPosOcc() == l.isPos()){++cnt;}else{++cntf;}
		}
	}
	bool ret = true;
	if(cnt != 1){
		LOG0("Variable of lit " << l.toDIMACS() << " contains cls " << e.idx << " " << cnt << " times");
		ret = false;
	}
	if(cntf != 0){
		LOG0("Variable of lit " << l.toDIMACS() << " contains cls " << e.idx << " in flipped polarity " << cntf << " times");
		ret = false;
	}
	return ret;
}

bool SimplyFire::asrtOccsContainCls( const ECP& e){
	ExtCls& ec = clses.getECls(e);
	if(ec.isWaste()){return true;}

	//LOG0("asrtOccsContainCls for wasted cls " << e.idx); return false;}

	Literal lit = ec.getLit();
	bool ok = asrtOccsContain(lit,e);

	Cls& cls	= glob.getCls(ec.getClp());
	ok 			= asrtOccsContain(cls.otherWatcher(lit),e) && ok;
	ClsLit* clit= cls.uwBeg();

	do{
		ok		= asrtOccsContain(Literal(clit->lit),e) && ok;
	}while((clit++)->has_next);

	return ok;
}

bool SimplyFire::asrtOccsContainAll(){
	bool ok = true;
	for(uint32_t c=clses.maxClses();c--;){
		ECP e(c);
		ok = asrtOccsContainCls(e) && ok;
	}
	if(ok){LOG2("All clses are linked in occs list ");}
	return ok;
}




