/*
 * Globals.hpp
 *
 *  Created on: Sep 29, 2010
 *      Author: kottlers
 */

#ifndef GLOBALS_HPP_
#define GLOBALS_HPP_
#define globals_hpp_version "$Revision: 1.15 $"

#include "Settings.hpp"
#include "Satistics.hpp"
#include "MemPoolA.hpp"
#include "MemPoolP.hpp"
#include "BitVector.hpp"
#include "Random.hpp"
#include "Cls.hpp"
#include "ElimClses.hpp"




struct TGlobals{
	Settings			sets;
	Satistics 			stats;
	MemPoolP			c_info;
	BitVector<Variable>	elim;
	ElimClses			elim_clses;
	char*				state_ok;
	Random				randy;

	uint32_t			curr_long_lrnts;
	uint32_t			lbd_distri[BDReason::LBD_MAX+1];


	bool			isElim(const Variable& var){ asrt(elim.size()>var.idx());
						return elim.isSet(var);}

	bool			isElim(const Literal& lit){return isElim(lit.toVar());}

	void			setElim(const Variable& var){ asrt(elim.size()>var.idx());
					LOG2("Set elim " << var.toDIMACS() );
						elim.setFlag(var);}

	Cls&		getCls(const ClsP& p) {
					return c_info.getRef<Cls>(p);}


	void			ensureVarVals(uint32_t sz){
						asrt(sz > elim.size());
						elim.ensureSize(sz,true);}

	bool 		setInvalid(const char* why){
					LOG1("Set invalid with reason " << std::string(why));
					if(state_ok==NULL){ state_ok = (char*)why;}
					return false;}

	bool 		isValid(){	return state_ok == NULL;}

	char* 		whyInvalid(){ asrt(state_ok!=NULL);
						return state_ok;}

	void 		registrLBD(const uint32_t& l){asrt(l<=BDReason::LBD_MAX);
						++lbd_distri[l];++curr_long_lrnts;}

	void 		unregistrLBD(const uint32_t& l){asrt(l<=BDReason::LBD_MAX); asrt(lbd_distri[l]>0);
					--lbd_distri[l]; --curr_long_lrnts;}

	void 		clearLBDDistri(){
					for(uint32_t i=BDReason::LBD_MAX+1;i--;){lbd_distri[i]=0;}
					curr_long_lrnts = 0;}

	TGlobals():state_ok(NULL),curr_long_lrnts( 0){clearLBDDistri();}

	inline ClsP clsCreate(Literal* lits, uint32_t sz, const bool learnt,const uint32_t& lbd);


	/*inline void	clsFree(const ClsP& clp);
	inline void	silentFree(const Ptr32& clp, uint32_t mem_sz);
	inline void 	clsShrink(const ClsP& clp,const uint32_t keep, const uint32_t rem);
	inline void 	silentShrink(const ClsP& clp,const uint32_t shrnk);*/
};


ClsP TGlobals::clsCreate(Literal* lits, uint32_t sz, const bool learnt,const uint32_t& lbd){

    asrt(sz > 2);

    ClsP loc	= c_info.alloc(Cls::allocSizeFor(sz));

    if(loc.isNull()){
    	cout << endl << "c !!!OUT OF MEMORY!!!" << endl;
    	exit(EXIT_FAILURE);
    	return ClsP();
    }

    uint32_t LBD	= lbd - (lbd!=0);
    new (c_info.getRealPtr<Cls>(loc)) 	Cls(lits[0],lits[1],learnt,LBD&1);
    {Cls& cls		= getCls(loc); asrt(learnt == cls.isLrnt() );
    ClsLit* it		= cls.uwBeg();

    for(uint32_t i = 2;i != sz;++i) {LBD>>=1; *it++ = ClsLit(lits[i],true,LBD&1);}
    (it-1)->has_next = 0; asrt(!learnt || cls.compLBD()==lbd);}

    stats.cnt_lng_lts += sz;

    return loc;
}



/*

void TGlobals::silentFree(const Ptr32& clp, uint32_t mem_sz){
	c_info			.dealloc(clp,mem_sz);}

void TGlobals::silentShrink(const ClsP& clp,const uint32_t shrnk){
	Cls& cls		= getCls(clp); asrt(cls.unWatchdCnt()>shrnk);
	cls			.reduceUnWatchdCount(shrnk);
	asrt(c_info.validEndData(clp)); asrt(c_info.validEndData(clp+(cls.memSize()-1)));
	c_info		.dealloc(clp+cls.memSize(),shrnk);}
*/

/*
void TGlobals::clsFree(const ClsP& clp){
	Cls& cls		= getCls(clp);
	uint32_t mem	= cls.computeMem();
	stats.nbr_lits -= (mem+1); //asrt(!cls.flagged());
	stats.nbr_lcls--;
	c_info			.dealloc(clp,mem);}


void TGlobals::clsShrink(const ClsP& clp,const uint32_t keep, const uint32_t rem){ asrt(rem>0); asrt(keep>0);
	Cls& cls		= getCls(clp);
	ClsLit* lits	= cls.uwBeg();
	stats			.nbr_lits -= rem;
	lits[keep-1]	.has_next = 0;
	asrt(c_info.validEndData(clp)); asrt(c_info.validEndData(clp+keep));
	c_info			.dealloc(clp+keep+1,rem);}
*/






#endif /* GLOBALS_HPP_ */
